/* 
 * Copyright (C) 2006 Darren Salt <linux@youmustbejoking.demon.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>

#include <gtk/gtkiconfactory.h>
#include <gtk/gtkicontheme.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkimage.h>
#include <gtk/gtkimagemenuitem.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <glib/gstdio.h>
                     
#include "gtk_icon_theme_context_menu.h"

// FIXME
#define _(x) (x)

static void gtk_icon_theme_context_menu_class_init (GtkIconThemeContextMenuClass []);
static void gtk_icon_theme_context_menu_init (GtkIconThemeContextMenu *);
static void gtk_icon_theme_context_menu_set_property (GObject *, guint,
						      const GValue *,
						      GParamSpec *);
static void gtk_icon_theme_context_menu_get_property (GObject *, guint,
						      GValue *, GParamSpec *);
static void gtk_icon_theme_context_menu_destroy (GtkObject *);

static void gtk_icon_theme_context_menu_maybe_update (GtkIconThemeContextMenu *);
static void gtk_icon_theme_context_menu_update (GtkIconThemeContextMenu *);

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), gtk_icon_theme_context_menu_get_type (), \
				GtkIconThemeContextMenuPrivate))

enum
{
  PROP_0,
  PROP_THEME,
  PROP_THEME_NAME,
  PROP_CONTEXT,
  PROP_LAST
};

typedef struct _GtkIconThemeContextMenuPrivate GtkIconThemeContextMenuPrivate;
struct _GtkIconThemeContextMenuPrivate
{
  gboolean destroyed;
  gboolean needs_update;
  GtkIconTheme *theme;
  gchar *context;
};

static GtkMenuClass *parent_class = NULL;

enum
{
  SIGNAL_ACTIVATED,
  SIGNAL_LAST
};

static guint signals[SIGNAL_LAST];

GType
gtk_icon_theme_context_menu_get_type (void)
{
  static GType type = 0;

  if (!type)
  {
    static const GTypeInfo info = {
      sizeof (GtkIconThemeContextMenuClass),
      NULL,			/* base_init */
      NULL,			/* base_finalize */
      (GClassInitFunc) gtk_icon_theme_context_menu_class_init,
      NULL,			/* class_finalize */
      NULL,			/* class_data */
      sizeof (GtkIconThemeContextMenu),
      0,			/* n_preallocs */
      (GInstanceInitFunc) gtk_icon_theme_context_menu_init,
      NULL			/* value_table */
    };

    type =
      g_type_register_static (GTK_TYPE_MENU, "GtkIconThemeContextMenu", &info,
			      0);
  }

  return type;
}

static void
gtk_icon_theme_context_menu_show (GtkWidget *widget)
{
  GtkIconThemeContextMenuPrivate *priv = GET_PRIVATE (widget);
  if (priv->needs_update)
    gtk_icon_theme_context_menu_update (GTK_ICON_THEME_CONTEXT_MENU (widget));
  (*GTK_WIDGET_CLASS (parent_class)->show) (widget);
}

static void
gtk_icon_theme_context_menu_class_init (GtkIconThemeContextMenuClass klass[])
{
  GObjectClass *const gobject_class = (GObjectClass *) klass;
  GtkObjectClass *const object_class = (GtkObjectClass *) klass;
  GtkWidgetClass *const widget_class = (GtkWidgetClass *) klass;

  parent_class = g_type_class_peek_parent (klass);

  object_class->destroy = gtk_icon_theme_context_menu_destroy;

  widget_class->show = gtk_icon_theme_context_menu_show;

  klass->activate = NULL;

  gobject_class->set_property = gtk_icon_theme_context_menu_set_property;
  gobject_class->get_property = gtk_icon_theme_context_menu_get_property;

  signals[SIGNAL_ACTIVATED] =
    g_signal_new ("activate",
		  G_TYPE_FROM_CLASS (object_class),
		  G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
		  G_STRUCT_OFFSET (GtkIconThemeContextMenuClass, activate),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__STRING,
		  G_TYPE_NONE, 1, G_TYPE_STRING);

  g_object_class_install_property
    (gobject_class, PROP_THEME,
     g_param_spec_object ("theme", _("Theme"),
			  _("The icon theme listed in the menu"),
			  gtk_icon_theme_context_menu_get_type (),
			  G_PARAM_READABLE | G_PARAM_WRITABLE));

  g_object_class_install_property
    (gobject_class, PROP_THEME_NAME,
     g_param_spec_string ("theme-name", _("Theme name"),
			  _("The name of the icon theme listed in the menu"),
			  NULL, G_PARAM_WRITABLE));

  g_object_class_install_property
    (gobject_class, PROP_CONTEXT,
     g_param_spec_string ("context", _("Context"),
			  _("The name of the context to be listed (NULL for all)"),
			  NULL, G_PARAM_READABLE | G_PARAM_WRITABLE));

  g_type_class_add_private (gobject_class,
			    sizeof (GtkIconThemeContextMenuPrivate));
}

static void
gtk_icon_theme_context_menu_init (GtkIconThemeContextMenu *menu)
{
  GtkIconThemeContextMenuPrivate *priv = GET_PRIVATE (menu);
  priv->destroyed = FALSE;
  priv->needs_update = TRUE;
  priv->theme = NULL;
  priv->context = NULL;
}

static void
gtk_icon_theme_context_menu_set_property (GObject *object, guint prop_id,
					  const GValue *value,
					  GParamSpec *pspec)
{
  GtkIconThemeContextMenu *menu = GTK_ICON_THEME_CONTEXT_MENU (object);

  switch (prop_id)
  {
  case PROP_THEME:
    gtk_icon_theme_context_menu_set_theme (menu, g_value_get_object (value));
    break;

  case PROP_THEME_NAME:
    gtk_icon_theme_context_menu_set_named_theme (menu,
						 g_value_get_string (value));
    break;

  case PROP_CONTEXT:
    gtk_icon_theme_context_menu_set_context (menu,
					     g_value_get_string (value));
    break;

  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gtk_icon_theme_context_menu_get_property (GObject *object, guint prop_id,
					  GValue *value, GParamSpec *pspec)
{
  GtkIconThemeContextMenu *menu = GTK_ICON_THEME_CONTEXT_MENU (object);
  GtkIconThemeContextMenuPrivate *priv = GET_PRIVATE (menu);

  switch (prop_id)
  {
  case PROP_THEME:
    g_value_set_object (value, priv->theme);
    break;

  case PROP_CONTEXT:
    g_value_set_string (value, priv->context);
    break;

  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gtk_icon_theme_context_menu_destroy (GtkObject *object)
{
  GtkIconThemeContextMenu *menu = GTK_ICON_THEME_CONTEXT_MENU (object);
  g_return_if_fail (menu != NULL);

  GtkIconThemeContextMenuPrivate *priv = GET_PRIVATE (menu);

  if (priv->destroyed)
    return;

  priv->destroyed = TRUE;

  g_object_unref (G_OBJECT (priv->theme));
  g_free (priv->context);

  (*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static GtkWidget *
_gtk_icon_theme_context_menu_new (GtkIconTheme *theme)
{
  GtkIconThemeContextMenu *menu =
    g_object_new (gtk_icon_theme_context_menu_get_type (), NULL);
  gtk_icon_theme_context_menu_set_theme (menu, theme);
  return GTK_WIDGET (menu);
}

/**
 * gtk_icon_theme_context_menu_new_with_default_theme:
 * @returns: A new #GtkIconThemeContextMenu.
 *
 * Creates a new #GtkIconThemeContextMenu with the default theme.
 **/
GtkWidget *
gtk_icon_theme_context_menu_new (void)
{
  return _gtk_icon_theme_context_menu_new (gtk_icon_theme_get_default ());
}

/**
 * gtk_icon_theme_context_menu_new_with_theme:
 * @returns: A new #GtkIconThemeContextMenu.
 *
 * Creates a new #GtkIconThemeContextMenu with a custom theme.
 **/
GtkWidget *
gtk_icon_theme_context_menu_new_with_theme (GtkIconTheme * theme)
{
  return _gtk_icon_theme_context_menu_new (theme);
}

/**
 * gtk_icon_theme_context_menu_new_with_named_theme:
 * @returns: A new #GtkIconThemeContextMenu.
 *
 * Creates a new #GtkIconThemeContextMenu with a custom theme.
 **/
GtkWidget *
gtk_icon_theme_context_menu_new_with_named_theme (const gchar *theme)
{
  GtkWidget *menu =
    gtk_icon_theme_context_menu_new_with_theme (gtk_icon_theme_new ());
  gtk_icon_theme_context_menu_set_named_theme
    ((GtkIconThemeContextMenu *) menu, theme);
  return menu;
}

/**
 * gtk_icon_theme_context_menu_set_theme:
 * @menu: An #GtkIconThemeContextMenu, created using #gtk_icon_theme_context_menu_new_with_theme().
 * @theme: Pointer to a GtkIconTheme object.
 *
 * Sets @theme as the theme whose icons are listed in the menu.
 **/
void
gtk_icon_theme_context_menu_set_theme (GtkIconThemeContextMenu *menu,
				       GtkIconTheme *theme)
{
  g_return_if_fail (GTK_IS_ICON_THEME_CONTEXT_MENU (menu));
  g_return_if_fail (GTK_IS_ICON_THEME (theme));
  GtkIconThemeContextMenuPrivate *priv = GET_PRIVATE (menu);
  if (priv->theme)
    g_object_unref (G_OBJECT (priv->theme));
  priv->theme = theme;
  g_object_ref (G_OBJECT (priv->theme));
  gtk_icon_theme_context_menu_maybe_update (menu);
}

/**
 * gtk_icon_theme_context_menu_set_named_theme:
 * @menu: An #GtkIconThemeContextMenu, created using #gtk_icon_theme_context_menu_new_with_theme().
 * @theme: The name of the icon theme.
 *
 * Sets @theme as the theme whose icons are listed in the menu.
 * Will fail if the menu is using a default icon theme object.
 **/
void
gtk_icon_theme_context_menu_set_named_theme (GtkIconThemeContextMenu *menu,
					     const gchar *theme)
{
  g_return_if_fail (GTK_IS_ICON_THEME_CONTEXT_MENU (menu));
  GtkIconThemeContextMenuPrivate *priv = GET_PRIVATE (menu);
  gtk_icon_theme_set_custom_theme (priv->theme, theme);
  gtk_icon_theme_context_menu_maybe_update (menu);
}

/**
 * gtk_icon_theme_context_menu_set_context:
 * @menu: An #GtkIconThemeContextMenu, created using #gtk_icon_theme_context_menu_new_with_theme().
 * @context: The icon theme context, determining which icons are listed.
 *
 * Sets @context as the context which controls which of the theme's icons
 * are listed in the menu.
 **/
void
gtk_icon_theme_context_menu_set_context (GtkIconThemeContextMenu *menu,
					 const gchar *context)
{
  g_return_if_fail (GTK_IS_ICON_THEME_CONTEXT_MENU (menu));

  GtkIconThemeContextMenuPrivate *priv = GET_PRIVATE (menu);
  g_free (priv->context);
  priv->context = context ? g_strdup (context) : NULL;
  gtk_icon_theme_context_menu_maybe_update (menu);
}

/**
 * gtk_icon_theme_context_menu_get_theme:
 * @app_menu_item: An #GtkIconThemeContextMenu.
 * @returns: An icon theme object, owned by the menu.
 *
 * Returns the current theme set for the #GtkIconThemeContextMenu.
 **/
GtkIconTheme *
gtk_icon_theme_context_menu_get_theme (GtkIconThemeContextMenu *menu)
{
  g_return_val_if_fail (GTK_IS_ICON_THEME_CONTEXT_MENU (menu), NULL);

  GtkIconThemeContextMenuPrivate *priv = GET_PRIVATE (menu);
  return priv->theme;
}

/**
 * gtk_icon_theme_context_menu_get_context:
 * @app_menu_item: An #GtkIconThemeContextMenu.
 * @returns: A context name string.
 *
 * Returns the current context set for the #GtkIconThemeContextMenu.
 **/
G_CONST_RETURN gchar *
gtk_icon_theme_context_menu_get_context (GtkIconThemeContextMenu *menu)
{
  g_return_val_if_fail (GTK_IS_ICON_THEME_CONTEXT_MENU (menu), NULL);

  GtkIconThemeContextMenuPrivate *priv = GET_PRIVATE (menu);
  return priv->context;
}

struct icon_data {
  gchar *name;
  GtkIconInfo *info;
};

static inline const gchar *
get_name (const struct icon_data data[])
{
  if (!data)
    return "";
  if (!data->info)
    return data->name;
  return gtk_icon_info_get_display_name (data->info) ? : data->name;
}

static int
gtk_icon_theme_context_menu_strcmp (const struct icon_data a[],
				    const struct icon_data b[])
{
  return strcmp (get_name (a), get_name (b));
}

static void
gtk_icon_theme_context_menu_activate (GtkMenuItem *item, gpointer data)
{
  g_signal_emit (G_OBJECT (GTK_WIDGET (item)->parent),
		 signals[SIGNAL_ACTIVATED], 0, data);
}

static void
gtk_icon_theme_context_menu_maybe_update (GtkIconThemeContextMenu *menu)
{
  if (GTK_WIDGET_MAPPED (GTK_WIDGET (menu)))
    gtk_icon_theme_context_menu_update (menu);
  else
  {
    GtkIconThemeContextMenuPrivate *priv = GET_PRIVATE (menu);
    priv->needs_update = TRUE;
  }
}

static void
gtk_icon_theme_context_menu_update (GtkIconThemeContextMenu *menu)
{
  GList *icon, *icons = gtk_container_get_children (GTK_CONTAINER (menu));

  /* flush the menu */
  for (icon = icons; icon; icon = g_list_next (icon))
    gtk_widget_destroy (icon->data);
    /* hmm, memory leak here? */
  g_list_free (icons);
 
  GtkIconThemeContextMenuPrivate *priv = GET_PRIVATE (menu);
  priv->needs_update = FALSE;

  icons = gtk_icon_theme_list_icons (priv->theme, priv->context);

  for (icon = icons; icon; icon = g_list_next (icon))
  {
    GtkIconInfo *info =
      gtk_icon_theme_lookup_icon (priv->theme, icon->data, 16,
				  GTK_ICON_LOOKUP_USE_BUILTIN);
    const gchar *file = gtk_icon_info_get_filename (info);
    struct stat st;
    if (file && (g_lstat (file, &st) || S_ISLNK (st.st_mode)))
    {
      gtk_icon_info_free (info);
      g_free (icon->data);
      icon->data = NULL;
    }
    else
    {
      struct icon_data *data = g_malloc (sizeof (struct icon_data));
      data->name = icon->data;
      data->info = info;
      icon->data = data;
    }
  }

  gtk_widget_push_composite_child ();

  icons = g_list_sort (icons, (GCompareFunc) gtk_icon_theme_context_menu_strcmp);
  gboolean added = FALSE;

  gint ix, iy;
  gtk_icon_size_lookup (GTK_ICON_SIZE_MENU, &ix, &iy);

  for (icon = icons; icon; icon = g_list_next (icon))
  {
    if (!icon->data)
      continue;

    struct icon_data *data = icon->data;
    GtkWidget *item =
      gtk_image_menu_item_new_with_label
	(gtk_icon_info_get_display_name (data->info) ? : data->name);
    const gchar *file = gtk_icon_info_get_filename (data->info);
    GtkWidget *image;
    if (file)
      image = gtk_image_new_from_pixbuf (gdk_pixbuf_new_from_file_at_size
					   (file, ix, iy, NULL));
    else
      image = gtk_image_new_from_icon_name (data->name, GTK_ICON_SIZE_MENU);
    gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
    gtk_icon_info_free (data->info);
    gtk_widget_show (item);
    gtk_widget_set_composite_name (item, "named_icon_menu_item");
    gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
    g_signal_connect_data (G_OBJECT (item), "activate",
			   G_CALLBACK (gtk_icon_theme_context_menu_activate),
			   data->name, (GClosureNotify) g_free, 0);
    g_free (data);
    added = TRUE;
  }

  if (!added)
  {
    GtkWidget *item = gtk_menu_item_new_with_label (_("(empty)"));
    gtk_widget_show (item);
    gtk_widget_set_sensitive (item, FALSE);
    gtk_widget_set_composite_name (item, "named_icon_menu_item");
    gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
  }

  gtk_widget_pop_composite_child ();

  g_list_free (icons);
}
