#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <gtk/gtk.h>

#include <gdk/gdkx.h>
#include <X11/Xlib.h>
#include <X11/cursorfont.h>

#include "gtk_icon_theme_menu.h"
#include "gtk_icon_theme_context_menu.h"

static int menu_x = -1, menu_y = -1;

static void
activated (GtkIconThemeMenu * menu G_GNUC_UNUSED, const gchar * icon,
	   gpointer data G_GNUC_UNUSED)
{
  printf ("%s%s", icon, isatty (1) ? "\n" : "");
  gtk_main_quit ();
}

static void
deactivated (GtkIconThemeMenu * menu G_GNUC_UNUSED,
	     gpointer data G_GNUC_UNUSED)
{
  gtk_main_quit ();
}

static void
position (GtkMenu * menu G_GNUC_UNUSED, gint * x, gint * y,
	  gboolean * push_in, gpointer data G_GNUC_UNUSED)
{
  if (menu_x != -1)
    *x = menu_x;
  if (menu_y != -1)
    *y = menu_y;
  *push_in = TRUE;
}

int
main (int argc, char **argv)
{
  GtkWidget *menu = NULL;
  int type = 0;
  const gchar *theme = NULL;

  gtk_init (&argc, &argv);

  int argp = 1;
  if (argc > argp && argv[argp][0] == '@')
  {
    int x, y;
    if (sscanf (argv[argp], "@%d,%d", &x, &y) == 2)
    {
      menu_x = x;
      menu_y = y;
    }
    ++argp;
  }

  if (argc > argp)
  {
    if (!strcmp (argv[argp], "--help") || !strcmp (argv[argp], "-h"))
    {
      printf ("Usage: %s [@X,Y] [.THEME] [+|-|= [CONTEXT ...]]\n"
	      "Displays a menu of named icons, and writes the selected item to stdout.\n",
	      argv[0]);
      return 0;
    }

    if (!strcmp (argv[argp], "--version") || !strcmp (argv[argp], "-v"))
    {
      printf ("%s " VERSION "\n"
	      "Copyright (C) 2006 Darren Salt\n"
	      "This is free software. You may redistribute copies of it under the terms of\n"
	      "the GNU General Public Licence <http://www.gnu.org/licenses/gpl.html>.\n"
	      "There is NO WARRANTY, to the extent permitted by law.",
	      argv[0]);
      return 0;
    }

    if (argv[argp][0] == '.')
      theme = argv[argp++] + 1;
  }

  GtkWidget *invisible = gtk_invisible_new ();
  GdkCursor *pointer = gdk_cursor_new (GDK_WATCH);
  gtk_widget_show (invisible);

  XGrabPointer (GDK_WINDOW_XDISPLAY (invisible->window),
		GDK_WINDOW_XID (invisible->window), True, 0,
		GrabModeAsync, GrabModeAsync, None,
		GDK_CURSOR_XCURSOR (pointer), CurrentTime);
                                   
  if (argc > argp)
  {
    if (!strcmp (argv[argp], "+"))
      type = ++argp, 1;
    else if (!strcmp (argv[argp], "-"))
      type = ++argp, 2;
    else if (!strcmp (argv[argp], "="))
    {
      ++argp;
      goto context_menu;
    }
    else
    {
      context_menu:
      if (argc == argp + 1)
      {
	menu = theme
	       ? gtk_icon_theme_context_menu_new_with_named_theme (theme)
	       : gtk_icon_theme_context_menu_new ();
	gtk_icon_theme_context_menu_set_context
	  (GTK_ICON_THEME_CONTEXT_MENU (menu), argv[argp]);
      }
      else
	type = 3;
    }

    if (type)
      menu = theme
	     ? gtk_icon_theme_menu_new_with_named_theme (theme)
	     : gtk_icon_theme_menu_new ();

    switch (type)
    {
    case 1:
      gtk_icon_theme_menu_add_contexts (GTK_ICON_THEME_MENU (menu),
					argv + argp);
      break;
    case 2:
      gtk_icon_theme_menu_remove_contexts (GTK_ICON_THEME_MENU (menu),
					   argv + argp);
      break;
    case 3:
      gtk_icon_theme_menu_set_contexts (GTK_ICON_THEME_MENU (menu),
					argv + argp);
      break;
    }
  }
  else
    menu = gtk_icon_theme_menu_new ();

  g_object_ref (G_OBJECT (menu));
  gtk_object_sink (GTK_OBJECT (menu));

  g_signal_connect (G_OBJECT (menu), "activate",
		    G_CALLBACK (activated), NULL);
  g_signal_connect (G_OBJECT (menu), "deactivate",
		    G_CALLBACK (deactivated), NULL);

  XUngrabPointer (GDK_WINDOW_XDISPLAY (invisible->window), CurrentTime);
  gtk_menu_popup (GTK_MENU (menu), NULL, NULL, position, NULL, 0,
		  GDK_CURRENT_TIME);

  gtk_main ();

  return 0;
}
